from astrologic.decorators.if_switcher import switcher as switcher  # noqa: F401
from astrologic.decorators.tail_recursion_optimization import no_recursion as no_recursion  # noqa: F401
from astrologic.decorators.inline import inline as inline  # noqa: F401
